@@ -12,6 +12,7 @@ import json
 import shutil
 import logging
 from itertools import product
+from functools import partial
 from tempfile import NamedTemporaryFile
 
 import six
@@ -20,7 +21,6 @@ from six.moves import urllib
 # Optional dependencies.
 try:
     import pyfits
-    pyfits = pyfits
 except ImportError:
     try:
         import astropy.io.fits as pyfits
@@ -29,22 +29,49 @@ except ImportError:
 
 try:
     import fitsio
-    fitsio = fitsio
 except ImportError:
     fitsio = None
 
 try:
     import numpy as np
-    np = np
 except ImportError:
     np = None
 
 try:
     import matplotlib.pyplot as pl
-    pl = pl
 except ImportError:
     pl = None
 
+try:
+    from tornado import gen
+    import tornado.ioloop
+    from tornado.httpclient import AsyncHTTPClient, HTTPRequest
+except ImportError:
+    AsyncHTTPClient = None
+
+    def async_download(*args, **kwargs):
+        raise ImportError("You need to install tornado to do an async "
+                          "download")
+else:
+    def handle_response(obj, response):
+        if response.error:
+            response.rethrow()
+        obj._save_fetched_file(response.body)
+
+    @gen.coroutine
+    def _async_download(lcs, clobber=False):
+        client = AsyncHTTPClient(max_clients=10)
+        to_fetch = [l for l in lcs if clobber or not l.cache_exists]
+        responses = yield [client.fetch(HTTPRequest(l.url))
+                           for l in to_fetch]
+        [handle_response(to_fetch[i], r) for i, r in enumerate(responses)]
+        raise gen.Return()
+
+    def async_download(lcs, clobber=False):
+        f = partial(_async_download, lcs, clobber=clobber)
+        ioloop = tornado.ioloop.IOLoop.instance()
+        ioloop.run_sync(f)
+
 from .config import KPLR_ROOT
 from . import mast
 
@@ -62,7 +89,7 @@ class API(object):
 
     ea_url = ("http://exoplanetarchive.ipac.caltech.edu/cgi-bin/nstedAPI"
               "/nph-nstedAPI")
-    mast_url = "http://archive.stsci.edu/kepler/{0}/search.php"
+    mast_url = "http://archive.stsci.edu/{0}/{1}/search.php"
 
     def __init__(self, data_root=None):
         self.data_root = data_root
@@ -126,7 +153,8 @@ class API(object):
 
         return [self._munge_dict(row) for row in result]
 
-    def mast_request(self, category, adapter=None, sort=None, **params):
+    def mast_request(self, category, adapter=None, sort=None, mission="kepler",
+                     **params):
         """
         Submit a request to the MAST API and return a dictionary of parameters.
 
@@ -153,7 +181,7 @@ class API(object):
 
         # Send the request.
         encoded_data = urllib.parse.urlencode(params).encode("ascii")
-        r = urllib.request.Request(self.mast_url.format(category),
+        r = urllib.request.Request(self.mast_url.format(mission, category),
                                    data=encoded_data)
         handler = urllib.request.urlopen(r)
         code = handler.getcode()
@@ -297,7 +325,36 @@ class API(object):
                              .format(kepid))
         return stars[0]
 
-    def _data_search(self, kepler_id=None, short_cadence=True, **params):
+    def k2_stars(self, **params):
+        """
+        Get a list of EPIC targets from MAST. Only return up to 100 results by
+        default.
+
+        :param **params:
+            The query parameters for the MAST API.
+
+        """
+        params["max_records"] = params.pop("max_records", 100)
+        stars = self.mast_request("epic", adapter=mast.epic_adapter,
+                                  mission="k2", **params)
+        return [K2Star(self, s) for s in stars]
+
+    def k2_star(self, id_):
+        """
+        Get a EPIC target by id from MAST.
+
+        :param id:
+            The integer ID of the star in the EPIC.
+
+        """
+        stars = self.k2_stars(id=id_, max_records=1)
+        if not len(stars):
+            raise ValueError("No EPIC target found with id: '{0}'"
+                             .format(id_))
+        return stars[0]
+
+    def _data_search(self, kepler_id=None, short_cadence=True,
+                     adapter=mast.dataset_adapter, **params):
         """
         Run a generic data search on MAST to return a list of dictionaries
         describing the data products.
@@ -318,13 +375,12 @@ class API(object):
         if not short_cadence:
             params["ktc_target_type"] = "LC"
 
-        data_list = self.mast_request("data_search",
-                                      adapter=mast.dataset_adapter,
+        data_list = self.mast_request("data_search", adapter=adapter,
                                       **params)
         return data_list
 
     def light_curves(self, kepler_id=None, short_cadence=True, fetch=False,
-                     clobber=False, **params):
+                     clobber=False, async=False, **params):
         """
         Find the set of light curves associated with a KIC target.
 
@@ -343,6 +399,9 @@ class API(object):
             A boolean flag that determines whether or not the data file should
             be overwritten even if it already exists.
 
+        :param async:
+            If ``True``, download the files asynchronously using Tornado.
+
         :param ** params:
             Other search parameters to be passed to the MAST data search.
 
@@ -350,11 +409,15 @@ class API(object):
         lcs = [LightCurve(self, d) for d in self._data_search(kepler_id,
                short_cadence=short_cadence, **params)]
         if fetch:
-            [l.fetch(clobber=clobber) for l in lcs]
+            if async:
+                async_download(lcs, clobber=clobber)
+            else:
+                [l.fetch(clobber=clobber) for l in lcs]
         return lcs
 
     def target_pixel_files(self, kepler_id=None, short_cadence=True,
-                           fetch=False, clobber=False, **params):
+                           fetch=False, clobber=False, async=False,
+                           cls=None, **params):
         """
         Find the set of target pixel files associated with a KIC target.
 
@@ -373,14 +436,22 @@ class API(object):
             A boolean flag that determines whether or not the data file should
             be overwritten even if it already exists.
 
+        :param async:
+            If ``True``, download the files asynchronously using Tornado.
+
         :param ** params:
             Other search parameters to be passed to the MAST data search.
 
         """
-        tpfs = [TargetPixelFile(self, d) for d in self._data_search(kepler_id,
+        if cls is None:
+            cls = TargetPixelFile
+        tpfs = [cls(self, d) for d in self._data_search(kepler_id,
                 short_cadence=short_cadence, **params)]
         if fetch:
-            [l.fetch(clobber=clobber) for l in tpfs]
+            if async:
+                async_download(tpfs, clobber=clobber)
+            else:
+                [l.fetch(clobber=clobber) for l in tpfs]
         return tpfs
 
 
@@ -443,50 +514,23 @@ class Model(object):
     def __repr__(self):
         return self.__str__()
 
-    def get_light_curves(self, short_cadence=True, fetch=False, clobber=False):
+    def get_light_curves(self, **kwargs):
         """
         Get a list of light curve datasets for the model and optionally
-        download the FITS files.
-
-        :param short_cadence:
-            A boolean flag that determines whether or not the short cadence
-            data should be included. (default: True)
-
-        :param fetch:
-            A boolean flag that determines whether or not the data file should
-            be downloaded.
-
-        :param clobber:
-            A boolean flag that determines whether or not the data file should
-            be overwritten even if it already exists.
+        download the FITS files. See :func:`API.light_curves` for parameter
+        documentation.
 
         """
-        return self.api.light_curves(self.kepid, short_cadence=short_cadence,
-                                     fetch=fetch, clobber=clobber)
+        return self.api.light_curves(self.kepid, **kwargs)
 
-    def get_target_pixel_files(self, short_cadence=True, fetch=False,
-                               clobber=False):
+    def get_target_pixel_files(self, **kwargs):
         """
         Get a list of target pixel datasets for the model and optionally
-        download the FITS files.
-
-        :param short_cadence:
-            A boolean flag that determines whether or not the short cadence
-            data should be included. (default: True)
-
-        :param fetch:
-            A boolean flag that determines whether or not the data file should
-            be downloaded.
-
-        :param clobber:
-            A boolean flag that determines whether or not the data file should
-            be overwritten even if it already exists.
+        download the FITS files. See :func:`API.target_pixel_files` for
+        parameter documentation.
 
         """
-        return self.api.target_pixel_files(self.kepid,
-                                           short_cadence=short_cadence,
-                                           fetch=fetch,
-                                           clobber=clobber)
+        return self.api.target_pixel_files(self.kepid, **kwargs)
 
 
 class KOI(Model):
@@ -580,24 +624,71 @@ class Star(Model):
         return self._kois
 
 
+class K2Star(Model):
+    """
+    A star from the `K2 EPIC Catalog (EPIC)
+    <http://archive.stsci.edu/search_fields.php?mission=epic>`_.
+
+    """
+
+    _id = "{id}"
+
+    def get_light_curves(self, **kwargs):
+        """
+        Get a list of light curve datasets for the model and optionally
+        download the FITS files. See :func:`API.light_curves` for parameter
+        documentation.
+
+        """
+        raise NotImplementedError("There aren't any light curves for K2 stars")
+
+    def get_target_pixel_files(self, **kwargs):
+        """
+        Get a list of target pixel datasets for the model and optionally
+        download the FITS files. See :func:`API.target_pixel_files` for
+        parameter documentation.
+
+        """
+        return self.api.target_pixel_files(ktc_k2_id=self.id, mission="k2",
+                                           adapter=mast.k2_dataset_adapter,
+                                           cls=K2TargetPixelFile,
+                                           **kwargs)
+
+
 class _datafile(Model):
 
     _id = "\"{sci_data_set_name}_{ktc_target_type}\""
-    base_url = "http://archive.stsci.edu/pub/kepler/{0}/{1}/{2}/{3}"
+
+    kepid_template = "{ktc_kepler_id:09d}"
+
     product = None
     suffixes = None
     filetype = None
 
     def __init__(self, *args, **params):
         super(_datafile, self).__init__(*args, **params)
-        self.kepid = "{0:09d}".format(int(self.ktc_kepler_id))
-        self.base_dir = os.path.join(self.api.data_root, "data", self.product,
-                                     self.kepid)
-
+        self.kepid = self.kepid_template.format(**(self.params))
         suffix = self.suffixes[int(self.ktc_target_type != "LC")]
         self._filename = "{0}_{1}{2}".format(self.sci_data_set_name,
                                              suffix, self.filetype).lower()
 
+    @classmethod
+    def local(cls, fn):
+        self = cls.__new__(cls)
+        self._name = "'local:{0}'".format(fn)
+        self.base_dir = ""
+        self._filename = fn
+        return self
+
+    @property
+    def base_dir(self):
+        """
+        The local base directory for these types of products.
+
+        """
+        return os.path.join(self.api.data_root, "data", self.product,
+                            self.kepid)
+
     @property
     def filename(self):
         """
@@ -613,8 +704,9 @@ class _datafile(Model):
         The remote URL for the data file on the MAST servers.
 
         """
-        return self.base_url.format(self.product, self.kepid[:4],
-                                    self.kepid, self._filename)
+        base_url = "http://archive.stsci.edu/pub/kepler/{0}/{1}/{2}/{3}"
+        return base_url.format(self.product, self.kepid[:4], self.kepid,
+                               self._filename)
 
     def open(self, clobber=False, **kwargs):
         """
@@ -657,6 +749,10 @@ class _datafile(Model):
         self.fetch(clobber=clobber)
         return fitsio.read(fn, **kwargs)
 
+    @property
+    def cache_exists(self):
+        return os.path.exists(self.filename)
+
     def fetch(self, clobber=False):
         """
         Download the data file from the server and save it locally. The local
@@ -669,7 +765,7 @@ class _datafile(Model):
         """
         # Check if the file already exists.
         filename = self.filename
-        if os.path.exists(filename) and not clobber:
+        if self.cache_exists and not clobber:
             logging.info("Found local file: '{0}'".format(filename))
             return self
 
@@ -681,7 +777,9 @@ class _datafile(Model):
         code = handler.getcode()
         if int(code) != 200:
             raise APIError(code, url, "")
+        return self._save_fetched_file(handler.read())
 
+    def _save_fetched_file(self, data):
         # Make sure that the root directory exists.
         try:
             os.makedirs(self.base_dir)
@@ -689,13 +787,14 @@ class _datafile(Model):
             pass
 
         # Save the contents of the file.
+        filename = self.filename
         logging.info("Saving file to: '{0}'".format(filename))
 
         # Atomically write to disk.
         # http://stackoverflow.com/questions/2333872/ \
         #        atomic-writing-to-file-with-python
         f = NamedTemporaryFile("wb", delete=False)
-        f.write(handler.read())
+        f.write(data)
         f.flush()
         os.fsync(f.fileno())
         f.close()
@@ -782,7 +881,7 @@ class TargetPixelFile(_datafile):
     suffixes = ["lpd-targ", "spd-targ"]
     filetype = ".fits.gz"
 
-    def plot(self):
+    def plot(self, normed=False):
         """
         Make a simple diagnostic plot of the target pixel light curves. The
         pixels in the optimal aperture are plotted in black and those outside
@@ -812,16 +911,91 @@ class TargetPixelFile(_datafile):
         # Loop over the pixels.
         m = np.isfinite(time)
         xlim = [np.min(time[m]), np.max(time[m])]
+        ylim = [0., 0.]
         for xi, yi in product(range(nx), range(ny)):
             ax = axes[xi, yi]
             f = flux[:, xi, yi]
 
             m = np.isfinite(f)
-            ax.plot(time[m], f[m], ".", color="rk"[int(aperture[xi, yi] == 3)],
-                    ms=2)
+            if normed:
+                if len(f[m]) > 0:
+                    ax.plot(time[m], (f[m]/np.median(f[m]))-1.0, ".",
+                            color="rk"[int(aperture[xi, yi] == 3)], ms=2)
+                    minval = np.where(
+                        np.min((f[m]/np.median(f[m]))-1.0) < ylim[0],
+                        np.min((f[m]/np.median(f[m]))-1.0),
+                        ylim[0])
+                    maxval = np.where(
+                        np.max((f[m]/np.median(f[m]))-1.0) > ylim[1],
+                        np.max((f[m]/np.median(f[m]))-1.0),
+                        ylim[1])
+                    minmax = [minval, maxval]
+            else:
+                ax.plot(time[m], f[m], ".",
+                        color="rk"[int(aperture[xi, yi] == 3)], ms=2)
 
             ax.set_xlim(xlim)
             ax.set_xticklabels([])
             ax.set_yticklabels([])
 
+        # If normed, put everything on the same scale
+        if normed:
+            for ax in axes.flatten():
+                try:
+                    ax.set_ylim(minmax)
+                except TypeError:
+                    continue
+
         return fig
+
+
+class K2TargetPixelFile(TargetPixelFile):
+    """
+    A reference to a K2 target pixel dataset on the MAST severs. Like the
+    :class:`LightCurve` object, this object handles local caching of the file
+    in a strict directory structure.
+
+    """
+
+    _id = "\"{sci_data_set_name}_{ktc_target_type}\""
+    kepid_template = "{ktc_k2_id:09d}"
+    product = "target_pixel_files"
+    suffixes = ["lpd-targ", "spd-targ"]
+    filetype = ".fits.gz"
+
+    @property
+    def base_dir(self):
+        """
+        The local base directory for these types of products.
+
+        """
+        return os.path.join(self.api.data_root, "data", "k2", self.product,
+                            self.kepid)
+
+    @property
+    def filename(self):
+        """
+        The local filename of the data file. This file is only guaranteed to
+        exist after ``fetch()`` has been called.
+
+        """
+        return os.path.join(self.base_dir, self._filename)
+
+    @property
+    def url(self):
+        """
+        The remote URL for the data file on the MAST servers.
+
+        """
+        if self.sci_campaign != 0:
+            raise NotImplementedError("Only campaign 0 is supported for now")
+        base_url = "http://archive.stsci.edu/pub/k2/"
+        if self.ktc_k2_id < 201000000:
+            base_url += "{0}/c0/200000000/{1:05d}/{2}"
+        else:
+            base_url += "{{0}}/c0/{0}/{{1:05d}}/{{2}}" \
+                .format(int(int(self.ktc_k2_id * 1e-5) * 1e5))
+
+        return base_url.format(self.product,
+                               int(int(int(self.kepid[-5:][-5:])*1e-3)*1e3),
+                               self._filename)
