@@ -8,7 +8,8 @@ Adapters for the field names/types returned by the MAST API.
 from __future__ import (division, print_function, absolute_import,
                         unicode_literals)
 
-__all__ = ["koi_adapter", "planet_adapter", "star_adapter", "dataset_adapter"]
+__all__ = ["koi_adapter", "planet_adapter", "star_adapter", "dataset_adapter",
+           "epic_adapter"]
 
 import logging
 
@@ -355,3 +356,133 @@ dataset_adapter = Adapter({
     "Skygroup_ID": ("sci_skygroup_id", int),
     "Condition flag": ("condition_flag", six.text_type),
 })
+
+epic_adapter = Adapter({
+    "EPIC": ("id", int),
+    "RA": ("k2_ra", float),
+    "Dec": ("k2_dec", float),
+    "KepMag": ("kp", float),
+    "HIP": ("hip", int),
+    "TYC": ("tyc", unicode),
+    "UCAC": ("ucac", unicode),
+    "2MASS": ("twomass", unicode),
+    "SDSS": ("sdss", unicode),
+    "Object type": ("objtype", unicode),
+    "Kepflag": ("kepflag", unicode),
+    "pmra": ("pmra", float),
+    "e_pmra": ("e_pmra", float),
+    "pmdec": ("pmdec", float),
+    "e_pmdec": ("e_pmdec", float),
+    "plx": ("plx", float),
+    "e_plx": ("e_plx", float),
+    "Bmag": ("bmag", float),
+    "e_Bmag": ("e_bmag", float),
+    "Vmag": ("vmag", float),
+    "e_Vmag": ("e_vmag", float),
+    "umag": ("umag", float),
+    "e_umag": ("e_umag", float),
+    "gmag": ("gmag", float),
+    "e_gmag": ("e_gmag", float),
+    "rmag": ("rmag", float),
+    "e_rmag": ("e_rmag", float),
+    "imag": ("imag", float),
+    "e_imag": ("e_imag", float),
+    "zmag": ("zmag", float),
+    "e_zmag": ("e_zmag", float),
+    "Jmag": ("jmag", float),
+    "e_Jmag": ("e_jmag", float),
+    "Hmag": ("hmag", float),
+    "e_Hmag": ("e_hmag", float),
+    "Kmag": ("kmag", float),
+    "e_Kmag": ("e_kmag", float),
+    "w1mag": ("w1mag", float),
+    "e_w1mag": ("e_w1mag", float),
+    "w2mag": ("w2mag", float),
+    "e_w2mag": ("e_w2mag", float),
+    "w3mag": ("w3mag", float),
+    "e_w3mag": ("e_w3mag", float),
+    "w4mag": ("w4mag", float),
+    "e_w4mag": ("e_w4mag", float),
+    "Teff": ("teff", float),
+    "e_teff": ("e_teff", float),
+    "logg": ("logg", float),
+    "e_logg": ("e_logg", float),
+    "[Fe/H]": ("feh", float),
+    "e_[Fe/H]": ("e_feh", float),
+    "Radius": ("rad", float),
+    "e_rad": ("e_rad", float),
+    "mass": ("mass", float),
+    "e_mass": ("e_mass", float),
+    "rho": ("rho", float),
+    "e_rho": ("e_rho", float),
+    "lum": ("lum", float),
+    "e_lum": ("e_lum", float),
+    "Distance": ("d", float),
+    "e_d": ("e_d", float),
+    "E(B-V)": ("ebv", float),
+    "2MASS Flag": ("mflg", unicode),
+    "Nearest Neighbor": ("prox", float),
+    "Nomad ID": ("nomad", unicode),
+})
+
+k2_dataset_adapter = Adapter({
+    "K2 ID": ("ktc_k2_id", int),
+    "Dataset Name": ("sci_data_set_name", unicode),
+    "Campaign": ("sci_campaign", int),
+    "Object type": ("objtype", unicode),
+    "Data Release": ("sci_data_rel", int),
+    "RA (J2000)": ("sci_ra", float),
+    "Dec (J2000)": ("sci_dec", float),
+    "Target Type": ("ktc_target_type", unicode),
+    "Archive Class": ("sci_archive_class", unicode),
+    "Ref": ("refnum", int),
+    "Actual Start Time": ("sci_start_time", unicode),
+    "Actual End Time": ("sci_end_time", unicode),
+    "Investigation ID": ("ktc_investigation_id", unicode),
+    "RA PM": ("pmRA", float),
+    "RA PM Err": ("e_pmRA", float),
+    "Dec PM": ("pmDEC", float),
+    "Dec PM Err": ("e_pmDEC", float),
+    "Plx": ("plx", float),
+    "Plx Err": ("e_plx", float),
+    "U Mag": ("umag", float),
+    "U Mag Err": ("e_umag", float),
+    "B Mag": ("bmag", float),
+    "B Mag Err": ("e_bmag", float),
+    "V Mag": ("vmag", float),
+    "V Mag Err": ("e_vmag", float),
+    "G Mag": ("gmag", float),
+    "G Mag Err": ("e_gmag", float),
+    "R Mag": ("rmag", float),
+    "R Mag Err": ("e_rmag", float),
+    "I Mag": ("imag", float),
+    "I Mag Err": ("e_imag", float),
+    "Z Mag": ("zmag", float),
+    "Z Mag Err": ("e_zmag", float),
+    "J Mag": ("jmag", float),
+    "J Mag Err": ("e_jmag", float),
+    "H Mag": ("hmag", float),
+    "H Mag Err": ("e_hmag", float),
+    "K Mag": ("kmag", float),
+    "K Mag Err": ("e_kmag", float),
+    "KEP Mag": ("kp", float),
+    "Kep Flag": ("kepflag", unicode),
+    "Hip ID": ("hip", int),
+    "Tyc ID": ("tyc", unicode),
+    "SDSS ID": ("sdss", unicode),
+    "UCAC ID": ("ucac", unicode),
+    "2MASS ID": ("twoMass", unicode),
+    "2MASS Flag": ("mflg", unicode),
+    "Processing Date": ("sci_generation_date", unicode),
+    "crowding": ("sci_crowdsap", float),
+    "contamination": ("sci_contamination", float),
+    "flux fraction": ("sci_flfrcsap", float),
+    "cdpp3": ("sci_Cdpp3_0", float),
+    "cdpp6": ("sci_Cdpp6_0", float),
+    "cdpp12": ("sci_Cdpp12_0", float),
+    "Module": ("sci_module", int),
+    "Output": ("sci_output", int),
+    "Channel": ("sci_channel", int),
+    "Nearest Neighbor": ("prox", float),
+    "Nomad ID": ("nomad", unicode),
+})
